//
//  ViewController.swift
//

import UIKit
import Speech

final class ViewController: UIViewController {

    // MARK: - IBOutlets

    @IBOutlet weak var inputTextView: UITextView!
    @IBOutlet weak var outputTextView: UITextView!
    
    @IBOutlet weak var recognizeButton: UIButton!
    @IBOutlet weak var resetButton: UIButton!

    // MARK: - Properties

    private let text: String = Constants.text

    private var recognizer: SpeechRecognizable = SpeechRecognizer()

    private var words: [Word] = []

    private var isRecognitionStarted: Bool = false {
        didSet {
            if isRecognitionStarted {
                recognizeButton.setTitle("Стоп", for: .normal)
                recognizeButton.backgroundColor = #colorLiteral(red: 0.7220714092, green: 0, blue: 0.1236564294, alpha: 1)
                recognizer.start()
            } else {
                recognizeButton.setTitle("Старт", for: .normal)
                recognizeButton.backgroundColor = #colorLiteral(red: 0.3411764801, green: 0.6235294342, blue: 0.1686274558, alpha: 1)
                recognizer.stop()
            }
        }
    }

    // MARK: - Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()
        configureView()

        parseText()

        recognizer.delegate = self
        recognizer.requestPermissions { [weak self] success in
            self?.recognizeButton.isEnabled = success
        }
    }

    // MARK: - Private methods

    private func configureView() {
        inputTextView.layer.cornerRadius = 5
        inputTextView.autocorrectionType = .no
        inputTextView.attributedText = text.attributed

        outputTextView.layer.cornerRadius = 5
        outputTextView.autocorrectionType = .no
        outputTextView.text = ""

        recognizeButton.layer.cornerRadius = 8
        recognizeButton.shadow()
        recognizeButton.isEnabled = false

        resetButton.layer.cornerRadius = 8
    }

    private func parseText() {
        words = text.splitToWords()
    }

    private func find(_ word: String, in words: [Word], threshold: Float = 0.80) -> [Word] {
        var matchedWords: [Word] = []
        for mainWord in words {
            var result = mainWord
            if word == mainWord.text {
                result.score = 1.0
                matchedWords.append(result)
            } else {
                let score = word.levenshteinScore(to: mainWord.text)
                if score >= threshold {
                    result.score = score
                    matchedWords.append(result)
                }
            }
        }
        return matchedWords
    }

    private func highlight(words: [Word]) {
        let attributed = NSMutableAttributedString(string: text)
        words.forEach {
            attributed.addAttribute(.backgroundColor, value: UIColor.green, range: NSRange(location: $0.offset, length: $0.text.count))
        }
        attributed.addAttribute(.font, value: UIFont.systemFont(ofSize: 14), range: NSRange(location: 0, length: text.count))
        inputTextView.attributedText = attributed
    }

    // MARK: - IBActions

    @IBAction func recognizeButtonTap(_ sender: Any) {
        isRecognitionStarted.toggle()
    }

    @IBAction func resetButtonTap(_ sender: Any) {
        inputTextView.attributedText = text.attributed
        outputTextView.text = ""
        parseText()
    }

    @IBAction func onScreenTap(_ sender: Any) {
        view.endEditing(true)
    }
}

extension ViewController: SpeechRecognizableDelegate {
    func output(result: SFSpeechRecognitionResult, isFinal: Bool) {
        guard !isFinal else {
            print("All readed text:", result.bestTranscription.formattedString)
            return
        }
        let words = result.bestTranscription.segments.compactMap({ $0.substring })
        print("Readed words:", words.joined(separator: ", "))
        DispatchQueue.main.async { [weak self] in
            guard let self = self, let word = words.last else {
                return
            }
            self.highlight(words: self.find(word, in: self.words))
        }
    }
}
